from collections import defaultdict
from typing import Dict, List, Mapping, MutableMapping, Optional, cast

from physrisk.data.zarr_reader import ZarrReader
from physrisk.kernel import calculation
from physrisk.kernel.hazards import Hazard, HazardKind

from ..kernel.hazard_model import (
    HazardDataRequest,
    HazardDataResponse,
    HazardEventDataResponse,
    HazardModel,
    HazardParameterDataResponse,
)
from .hazard_data_provider import AcuteHazardDataProvider, ChronicHazardDataProvider, HazardDataProvider, SourcePath


class PregeneratedHazardModel(HazardModel):
    """Hazard event model that processes requests using EventProviders."""

    def __init__(
        self,
        hazard_data_providers: Dict[type, HazardDataProvider],
    ):

        self.acute_hazard_data_providers = dict(
            (k, cast(AcuteHazardDataProvider, v))
            for (k, v) in hazard_data_providers.items()
            if Hazard.kind(k) == HazardKind.acute
        )
        self.chronic_hazard_data_providers = dict(
            (k, cast(ChronicHazardDataProvider, v))
            for (k, v) in hazard_data_providers.items()
            if Hazard.kind(k) == HazardKind.chronic
        )

    def get_hazard_events(self, requests: List[HazardDataRequest]) -> Mapping[HazardDataRequest, HazardDataResponse]:

        batches = defaultdict(list)
        for request in requests:
            batches[request.group_key()].append(request)

        responses: MutableMapping[HazardDataRequest, HazardDataResponse] = {}
        for key in batches.keys():
            batch: List[HazardDataRequest] = batches[key]
            event_type, model, scenario, year = batch[0].hazard_type, batch[0].model, batch[0].scenario, batch[0].year
            longitudes = [req.longitude for req in batch]
            latitudes = [req.latitude for req in batch]
            if event_type.kind == HazardKind.acute:  # type: ignore
                intensities, return_periods = self.acute_hazard_data_providers[event_type].get_intensity_curves(
                    longitudes, latitudes, model=model, scenario=scenario, year=year
                )

                for i, req in enumerate(batch):
                    responses[req] = HazardEventDataResponse(return_periods, intensities[i, :])
            elif event_type.kind == HazardKind.chronic:  # type: ignore
                parameters = self.chronic_hazard_data_providers[event_type].get_parameters(
                    longitudes, latitudes, model=model, scenario=scenario, year=year
                )

                for i, req in enumerate(batch):
                    responses[req] = HazardParameterDataResponse(parameters[i])
        return responses


class ZarrHazardModel(PregeneratedHazardModel):
    def __init__(
        self,
        source_paths: Optional[Dict[type, SourcePath]] = None,
        store=None,
        interpolation="floor",
    ):
        if source_paths is None:
            source_paths = calculation.get_default_zarr_source_paths()

        # share ZarrReaders across HazardDataProviders
        zarr_reader = ZarrReader(store=store)

        super().__init__(
            dict(
                (
                    t,
                    (
                        AcuteHazardDataProvider(sp, zarr_reader=zarr_reader, interpolation=interpolation)
                        if Hazard.kind(t) == HazardKind.acute
                        else ChronicHazardDataProvider(sp, zarr_reader=zarr_reader, interpolation=interpolation)
                    ),
                )
                for t, sp in source_paths.items()
            )
        )



# import numpy.typing as npt
from abc import ABC, abstractmethod


class AssetImpact:
    """Calculates the impacts associated with a portfolio of assets."""

    def __init__(self, assets, vulnerabilities):
        pass


class AssetEventProvider(ABC):
    @abstractmethod
    def get_asset_events(self, assets, event_types):
        """Source event distributions in the locale of each asset for events of certain types"""


class ModelsBuilder(ABC):
    """Provides VulnerabilityModels and EventProviders for a type of aset."""

    @abstractmethod
    def get_vulnerability_model(self, asset_type):
        pass

    @abstractmethod
    def get_event_data_provider(self, asset_type):
        """Return a list of backends matching the specified filtering.
        Args:
            asset_type (AssetType): type of asset.
        Returns:
            dict[EventType, AssetEvents]: a list of Backends that match the filtering
                criteria.
        """
        pass



from typing import List, Optional

import numpy as np
from pydantic import BaseModel, Field


class TypedArray(np.ndarray):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        return np.array(val, dtype=cls.inner_type)  # type: ignore


class ArrayMeta(type):
    def __getitem__(self, t):
        return type("Array", (TypedArray,), {"inner_type": t})


class Array(np.ndarray, metaclass=ArrayMeta):
    pass


class Asset(BaseModel):
    """Defines an asset. An asset is identified first by its asset_class and then by its type within the class.
    An asset's value may be impacted through damage or through disruption
    disruption being reduction of an asset's ability to generate cashflows
    (or equivalent value, e.g. by reducing expenses or increasing sales).
    """

    asset_class: str = Field(
        description="name of asset class; corresponds to physrisk class names, e.g. PowerGeneratingAsset"
    )
    type: Optional[str] = Field(description="Type of the asset <level_1>/<level_2>/<level_3>")
    location: Optional[str]
    latitude: float = Field(description="Latitude in degrees")
    longitude: float = Field(description="Longitude in degrees")


class Assets(BaseModel):
    """Defines a collection of assets."""

    items: List[Asset]


class Country(BaseModel):
    """Country information."""

    country: str
    continent: str
    country_iso_a3: str


class Countries(BaseModel):
    """List of Country."""

    items: List[Country]


class IntensityCurve(BaseModel):
    """Intensity curve of an acute hazard."""

    intensities: List[float]
    return_periods: List[float]


class ExceedanceCurve(BaseModel):
    """General exceedance curve (e.g. hazazrd, impact)."""

    values: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")
    exceed_probabilities: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")

    class Config:
        arbitrary_types_allowed = True


class Distribution(BaseModel):
    """General exceedance curve (e.g. hazazrd, impact)."""

    bin_edges: np.ndarray = Field(default_factory=lambda: np.zeros(11), description="")
    probabilities: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")

    class Config:
        arbitrary_types_allowed = True


class HazardEventDistrib(BaseModel):
    """Intensity curve of an acute hazard."""

    intensity_bin_edges: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")
    probabilities: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")

    class Config:
        arbitrary_types_allowed = True


class VulnerabilityCurve(BaseModel):
    """Defines a damage or disruption curve."""

    asset_type: str = Field(...)
    location: str = Field(...)
    event_type: str = Field(description="hazard event type, e.g. RiverineInundation")
    impact_type: str = Field(description="'Damage' or 'Disruption'")
    # intensity: Array = Field(...)
    # intensity: np.ndarray = np.zeros(1) #Field(default_factory=lambda: np.zeros(1))
    intensity: List[float] = Field(...)
    intensity_units: str = Field(description="units of the intensity")
    impact_mean: List[float] = Field(description="mean impact (damage or disruption)")
    impact_std: List[float] = Field(description="standard deviation of impact (damage or disruption)")

    class Config:
        arbitrary_types_allowed = True


class VulnerabilityCurves(BaseModel):
    """List of VulnerabilityCurve."""

    items: List[VulnerabilityCurve]


class VulnerabilityDistrib(BaseModel):
    """Defines a vulnerability matrix."""

    intensity_bin_edges: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")
    impact_bin_edges: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")
    prob_matrix: np.ndarray = Field(default_factory=lambda: np.zeros(10), description="")

    class Config:
        arbitrary_types_allowed = True



import importlib
import json
from importlib import import_module
from typing import Dict, List, cast

import numpy as np

import physrisk.data.static.example_portfolios
from physrisk.api.v1.common import Distribution, ExceedanceCurve, VulnerabilityDistrib

from .api.v1.hazard_data import (
    HazardEventAvailabilityRequest,
    HazardEventAvailabilityResponse,
    HazardEventDataRequest,
    HazardEventDataResponse,
    HazardEventDataResponseItem,
    IntensityCurve,
)
from .api.v1.impact_req_resp import (
    AcuteHazardCalculationDetails,
    AssetImpactRequest,
    AssetImpactResponse,
    AssetLevelImpact,
    Assets,
    AssetSingleHazardImpact,
)
from .data.inventory import Inventory
from .data.pregenerated_hazard_model import ZarrHazardModel
from .kernel import Asset, Hazard
from .kernel import calculation as calc
from .kernel.hazard_model import HazardDataRequest
from .kernel.hazard_model import HazardEventDataResponse as hmHazardEventDataResponse
from .kernel.hazard_model import HazardParameterDataResponse


class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def dumps(dict):
    return json.dumps(dict, cls=NumpyArrayEncoder)


def get(*, request_id, request_dict, store=None):

    if request_id == "get_hazard_data":
        request = HazardEventDataRequest(**request_dict)
        return json.dumps(_get_hazard_data(request, store=store).dict())
    elif request_id == "get_hazard_data_availability":
        request = HazardEventAvailabilityRequest(**request_dict)
        return json.dumps(_get_hazard_data_availability(request).dict())
    elif request_id == "get_asset_impact":
        request = AssetImpactRequest(**request_dict)
        return dumps(_get_asset_impacts(request, store=store).dict())
    elif request_id == "get_example_portfolios":
        return dumps(_get_example_portfolios())
    else:
        raise ValueError(f"request type '{request_id}' not found")


def _get_hazard_data_availability(request: HazardEventAvailabilityRequest):
    inventory = Inventory()
    # models = inventory.models
    models = inventory.to_hazard_models()
    colormaps = inventory.colormaps()
    response = HazardEventAvailabilityResponse(models=models, colormaps=colormaps)  # type: ignore
    return response


def _get_hazard_data(request: HazardEventDataRequest, source_paths=None, store=None):
    hazard_model = _create_hazard_model(interpolation=request.interpolation, source_paths=source_paths, store=store)

    # get hazard event types:
    event_types = Hazard.__subclasses__()
    event_dict = dict((et.__name__, et) for et in event_types)
    event_dict.update((est.__name__, est) for et in event_types for est in et.__subclasses__())

    # flatten list to let event processer decide how to group
    item_requests = []
    all_requests = []
    for item in request.items:
        event_type = event_dict[item.event_type]

        data_requests = [
            HazardDataRequest(event_type, lon, lat, model=item.model, scenario=item.scenario, year=item.year)
            for (lon, lat) in zip(item.longitudes, item.latitudes)
        ]

        all_requests.extend(data_requests)
        item_requests.append(data_requests)

    response_dict = hazard_model.get_hazard_events(all_requests)
    # responses comes back as a dictionary because requests may be executed in different order to list
    # to optimise performance.

    response = HazardEventDataResponse(items=[])

    for i, item in enumerate(request.items):
        requests = item_requests[i]
        # resps = (cast(hmHazardEventDataResponse, response_dict[req]) for req in requests)
        resps = (response_dict[req] for req in requests)
        intensity_curves = [
            IntensityCurve(intensities=list(resp.intensities), return_periods=list(resp.return_periods))
            if isinstance(resp, hmHazardEventDataResponse)
            else IntensityCurve(intensities=[resp.parameter], return_periods=[])
            if isinstance(resp, HazardParameterDataResponse)
            else None
            for resp in resps
        ]
        response.items.append(
            HazardEventDataResponseItem(
                intensity_curve_set=intensity_curves,
                request_item_id=item.request_item_id,
                event_type=item.event_type,
                model=item.model,
                scenario=item.scenario,
                year=item.year,
            )
        )

    return response


def create_assets(assets: Assets):
    """Create list of Asset objects from the Assets API object:"""
    module = import_module("physrisk.kernel.assets")
    asset_objs = []
    for asset in assets.items:
        asset_obj = cast(
            Asset,
            getattr(module, asset.asset_class)(
                asset.latitude, asset.longitude, type=asset.type, location=asset.location
            ),
        )
        asset_objs.append(asset_obj)
    return asset_objs


def _create_hazard_model(interpolation="floor", source_paths=None, store=None):
    if source_paths is None:
        source_paths = calc.get_default_zarr_source_paths()

    hazard_model = ZarrHazardModel(source_paths, store=store, interpolation=interpolation)

    return hazard_model


def _get_asset_impacts(request: AssetImpactRequest, source_paths=None, store=None):
    hazard_model = _create_hazard_model(source_paths=source_paths, store=store)
    vulnerability_models = calc.get_default_vulnerability_models()

    # we keep API definition of asset separate from internal Asset class; convert by reflection
    # based on asset_class:
    assets = create_assets(request.assets)

    results = calc.calculate_impacts(
        assets, hazard_model, vulnerability_models, scenario=request.scenario, year=request.year
    )

    # note that this does rely on ordering of dictionary (post 3.6)
    impacts: Dict[Asset, List[AssetSingleHazardImpact]] = {}
    for (asset, hazard_type), v in results.items():
        # calculation details
        if v.event is not None and v.vulnerability is not None:
            hazard_exceedance = v.event.to_exceedance_curve()

            vulnerability_distribution = VulnerabilityDistrib(
                intensity_bin_edges=v.vulnerability.intensity_bins,
                impact_bin_edges=v.vulnerability.impact_bins,
                prob_matrix=v.vulnerability.prob_matrix,
            )

            calc_details = AcuteHazardCalculationDetails(
                hazard_exceedance=ExceedanceCurve(
                    values=hazard_exceedance.values, exceed_probabilities=hazard_exceedance.probs
                ),
                hazard_distribution=Distribution(bin_edges=v.event.intensity_bin_edges, probabilities=v.event.prob),
                vulnerability_distribution=vulnerability_distribution,
            )

        impact_exceedance = v.impact.to_exceedance_curve()
        hazard_impacts = AssetSingleHazardImpact(
            hazard_type=v.impact.hazard_type.__name__,
            impact_type=v.impact.impact_type.name,
            impact_exceedance=ExceedanceCurve(
                values=impact_exceedance.values, exceed_probabilities=impact_exceedance.probs
            ),
            impact_distribution=Distribution(bin_edges=v.impact.impact_bins, probabilities=v.impact.prob),
            impact_mean=v.impact.mean_impact(),
            impact_std_deviation=0,  # TODO!
            calc_details=None if v.event is None else calc_details,
        )

        impacts.setdefault(asset, []).append(hazard_impacts)

    asset_impacts = [AssetLevelImpact(asset_id="", impacts=a) for a in impacts.values()]

    return AssetImpactResponse(asset_impacts=asset_impacts)


def _get_example_portfolios() -> List[Assets]:
    portfolios = []
    for file in importlib.resources.contents(physrisk.data.static.example_portfolios):
        if not str(file).endswith(".json"):
            continue
        with importlib.resources.open_text(physrisk.data.static.example_portfolios, file) as f:
            portfolio = Assets(**json.load(f))
            portfolios.append(portfolio)
    return portfolios



import importlib.util
import sys
from types import ModuleType


def lazy_import(name):
    spec = importlib.util.find_spec(name)
    if spec is not None:
        spec_loader = spec.loader
        assert spec_loader is not None
        loader = importlib.util.LazyLoader(spec_loader)
        spec.loader = loader
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        loader.exec_module(module)
        return module
    return ModuleType("not found", None)



from enum import Enum
from typing import cast


class HazardKind(Enum):
    acute = (1,)
    chronic = 2


class InundationType(Enum):
    riverine = (1,)
    coastal = 2


class Hazard:
    @staticmethod
    def kind(hazard_type):
        return cast(HazardKind, hazard_type.kind)


class Drought(Hazard):
    pass


class ChronicHeat(Hazard):
    kind = HazardKind.chronic
    pass


class Inundation(Hazard):
    kind = HazardKind.acute
    pass


class RiverineInundation(Inundation):
    kind = HazardKind.acute
    pass


class CoastalInundation(Inundation):
    kind = HazardKind.acute
    pass



from typing import List, Union

import numpy as np

from . import curve


class HazardEventDistrib:
    """Intensity distribution of a hazard event (e.g. inundation depth, wind speed etc),
    specific to an asset -- that is, at the location of the asset."""

    __slots__ = ["__event_type", "__intensity_bins", "__prob", "__exceedance"]

    def __init__(
        self, event_type: type, intensity_bins: Union[List[float], np.ndarray], prob: Union[List[float], np.ndarray]
    ):
        """Create a new asset event distribution.
        Args:
            event_type: type of event
            intensity_bins: non-decreasing intensity bin edges.
            e.g. bin edges [1.0, 1.5, 2.0] imply two bins: 1.0 < i <= 1.5, 1.5 < i <= 2.0
            prob: (annual) probability of occurrence for each intensity bin with size [len(intensity_bins) - 1]
        """
        self.__event_type = event_type
        self.__intensity_bins = np.array(intensity_bins)
        self.__prob = np.array(prob)

    def intensity_bins(self):
        return zip(self.__intensity_bins[0:-1], self.__intensity_bins[1:])

    def to_exceedance_curve(self):
        return curve.to_exceedance_curve(self.__intensity_bins, self.__prob)

    @property
    def intensity_bin_edges(self) -> np.ndarray:
        return self.__intensity_bins

    @property
    def prob(self) -> np.ndarray:
        return self.__prob



from typing import Iterable, Union

import numpy as np

from ..kernel.assets import Asset, PowerGeneratingAsset
from ..kernel.curve import ExceedanceCurve
from ..kernel.hazard_event_distrib import HazardEventDistrib
from ..kernel.hazard_model import HazardDataRequest, HazardDataResponse, HazardEventDataResponse
from ..kernel.hazards import ChronicHeat, RiverineInundation
from ..kernel.vulnerability_distrib import VulnerabilityDistrib
from ..kernel.vulnerability_model import (
    DeterministicVulnerabilityModel,
    VulnerabilityModelAcuteBase,
    applies_to_assets,
    applies_to_events,
)


@applies_to_events([RiverineInundation])
@applies_to_assets([PowerGeneratingAsset])
class InundationModel(VulnerabilityModelAcuteBase):
    def __init__(self, model="MIROC-ESM-CHEM"):
        # default impact curve
        self.__curve_depth = np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 1])
        self.__curve_impact = np.array([0, 1, 2, 7, 14, 30, 60, 180, 365])
        self.__model = model
        self.__base_model = "000000000WATCH"
        super().__init__(model, RiverineInundation)
        pass

    def get_data_requests(
        self, asset: Asset, *, scenario: str, year: int
    ) -> Union[HazardDataRequest, Iterable[HazardDataRequest]]:
        """Provide the list of hazard event data requests required in order to calculate
        the VulnerabilityDistrib and HazardEventDistrib for the asset."""

        histo = HazardDataRequest(
            RiverineInundation,
            asset.longitude,
            asset.latitude,
            scenario="historical",
            year=1980,
            model=self.__base_model,
        )

        future = HazardDataRequest(
            RiverineInundation, asset.longitude, asset.latitude, scenario=scenario, year=year, model=self.__model
        )

        return histo, future

    def get_distributions(self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]):
        """Return distributions for asset, based on hazard event date:
        VulnerabilityDistrib and HazardEventDistrib."""

        histo, future = event_data_responses
        assert isinstance(histo, HazardEventDataResponse)
        assert isinstance(future, HazardEventDataResponse)

        protection_return_period = 250.0
        curve_histo = ExceedanceCurve(1.0 / histo.return_periods, histo.intensities)
        # the protection depth is the 250-year-return-period inundation depth at the asset location
        protection_depth = curve_histo.get_value(1.0 / protection_return_period)

        curve_future = ExceedanceCurve(1.0 / future.return_periods, future.intensities)
        curve_future = curve_future.add_value_point(protection_depth)

        depth_bins, probs = curve_future.get_probability_bins()

        impact_bins = np.interp(depth_bins, self.__curve_depth, self.__curve_impact) / 365.0

        # keep all bins, but make use of vulnerability matrix to apply protection level
        # for improved performance we could truncate (and treat identify matrix as a special case)
        # but this general version allows model uncertainties to be added
        probs_protected = np.where(depth_bins[1:] <= protection_depth, 0.0, 1.0)

        vul = VulnerabilityDistrib(RiverineInundation, depth_bins, impact_bins, np.diag(probs_protected))
        event = HazardEventDistrib(RiverineInundation, depth_bins, probs)

        return vul, event


@applies_to_events([ChronicHeat])
@applies_to_assets([PowerGeneratingAsset])
class TemperatureModel(DeterministicVulnerabilityModel):
    def __init__(self):
        # does nothing
        pass



from typing import List, Optional

from pydantic import BaseModel, Field

from physrisk.api.v1.common import Assets, Distribution, ExceedanceCurve, VulnerabilityDistrib

# region Request


class CalcSettings(BaseModel):
    hazard_interp: str = Field("floor", description="Method used for interpolation of hazards: 'lloor' or 'bilinear'.")


class AssetImpactRequest(BaseModel):
    """Impact calculation request."""

    assets: Assets
    calc_settings: CalcSettings = Field(default_factory=CalcSettings, description="Interpolation method.")
    include_asset_level: bool = Field(True, description="If true, include ")
    include_calc_details: bool = Field(True, description="If true, include impact calculation details.")
    scenario: str = Field("rcp8p5", description="Name of scenario ('rcp8p5')")
    year: int = Field(
        2050,
        description="Projection year (2030, 2050, 2080). Any year before 2030, e.g. 1980, is treated as historical.",
    )


# endregion

# region Response


class AcuteHazardCalculationDetails(BaseModel):
    """Details of an acute hazard calculation."""

    hazard_exceedance: ExceedanceCurve
    hazard_distribution: Distribution
    vulnerability_distribution: VulnerabilityDistrib


class AssetSingleHazardImpact(BaseModel):
    """Impact at level of single asset and single type of hazard."""

    hazard_type: str = Field("", description="Type of the hazard.")
    impact_type: str = Field(
        "damage",
        description="""'damage' or 'disruption'. Whether the impact is fractional damage to the asset
        ('damage') or disruption to the annual economic benefit obtained from the asset ('disruption'), expressed as
        fractional decrease to an equivalent cash amount.""",
    )
    impact_distribution: Distribution
    impact_exceedance: ExceedanceCurve
    impact_mean: float
    impact_std_deviation: float
    calc_details: Optional[AcuteHazardCalculationDetails] = Field(
        None,
        description="""Details of impact calculation for acute hazard calculations.""",
    )

    class Config:
        arbitrary_types_allowed = True


class AssetLevelImpact(BaseModel):
    """Impact at asset level. Each asset can have impacts for multiple hazard types."""

    asset_id: Optional[str] = Field(
        None,
        description="""Asset identifier; will appear if provided in the request
        otherwise order of assets in response is identical to order of assets in request.""",
    )
    impacts: List[AssetSingleHazardImpact] = Field([], description="Impacts for each hazard type.")


class AssetImpactResponse(BaseModel):
    """Response to impact request."""

    asset_impacts: List[AssetLevelImpact]


# endregion



# flake8: noqa: E501
import hashlib
from typing import List

from pydantic import parse_obj_as

import physrisk.data.colormap_provider as colormap_provider

from ..api.v1.hazard_data import HazardModel, Period

methodology_doc = """
For more details and relevant citations see the
[OS-Climate Physical Climate Risk Methodology document](https://github.com/os-climate/physrisk/blob/main/methodology/PhysicalRiskMethodology.pdf).
"""

aqueduct_description = (
    """
The World Resources Institute (WRI) [Aqueduct Floods model](https://www.wri.org/aqueduct) is an acute riverine and
coastal flood hazard model with a spatial resolution of 30 × 30 arc seconds (approx. 1 km at the equator). Flood intensity is provided
as a _return period_ map: each point comprises a curve of inundation depths for 9 different return periods (also known as
reoccurrence periods). If a flood event has depth $d_i$ with return period of $r_i$ this implies that the probability of a flood
event with depth greater than $d_i$ occurring in any one year is $1 / r_i$; this is the _exceedance probability_.

Aqueduct Floods is based on Global Flood Risk with IMAGE Scenarios (GLOFRIS); see [here](https://www.wri.org/aqueduct/publications)
for more details.

"""
    + methodology_doc
)


class Inventory:
    """Contains an inventory of available hazard data.
    model id is given by {event_type}/{model group identifier}/{version}/{model identifier}
    """

    def __init__(self):
        osc_chronic_heat_models = [
            {
                "event_type": "ChronicHeat",
                "path": "",
                "id": "mean_degree_days/above/32c",
                "display_name": "Mean degree days above 32°C",
                "description": """
Degree days indicators are calculated by integrating over time the absolute difference in temperature
of the medium over a reference temperature. The exact method of calculation may vary;
here the daily average temperature is used to calculate an annual indicator:
$$
I^\\text{dd} = \\sum_{i = 1}^{365} |  T^\\text{avg}_i - T^\\text{ref} | \\nonumber
$$
$I^\\text{dd}$ is the indicator, $T^\\text{avg}$ is the daily average surface temperature
and $T^\\text{ref}$ is the reference temperature of 32°C. The OS-Climate-generated indicators are inferred
from CMIP6 data, averaged over 6 models: ACCESS-CM2, CMCC-ESM2, CNRM-CM6-1, MPI-ESM1-2-LR, MIROC6 and NorESM2-MM.
The indicators are generated for periods: 'historical' (averaged over 1995-2014), 2030 (2021-2040), 2040 (2031-2050)
and 2050 (2041-2060).

Applications for indicators based on surface temperature degree days include models of:
- heating and cooling requirements
- labour loss caused by extreme heat

                """
                + methodology_doc,
                "filename": "mean_degree_days_above_32c_{scenario}_{year}",
                "map": {"colormap": "mean_degree_days/above/32c"},
                "units": "degree days",
                "scenarios": [
                    {"id": "ssp585", "years": [2030, 2040, 2050]},
                    {"id": "historical", "years": [1980]},
                ],
            },
            {
                "event_type": "ChronicHeat",
                "path": "",
                "id": "mean_work_loss/high",
                "display_name": "Mean work loss",
                "description": """
The mean work loss indicator is calculated from the 'Wet Bulb Globe Temperature' (WBGT) indicator:
$$
I^\\text{WBGT}_i = 0.567 \\times T^\\text{avg}_i + 0.393 \\times P^\\text{vapour}_i
$$
$I^\\text{WBGT}_i$ is the WBGT indicator, $T^\\text{avg}_i$ is the daily average surface temperature (in degress Celsius) on day index, $i$, and $P^\\text{vapour}$
is the water vapour partial pressure (in kPa). $P^\\text{vapour}$ is calculated from relative humidity $H_R$ via:
$$
P^\\text{vapour}_i = \\frac{H_R}{100} \\times 6.105 \\times \\exp \\left( \\frac{17.27 \\times T^\\text{avg}_i}{237.7 \\times T^\\text{avg}_i} \\right)
$$
The work ability indicator, $I^{\\text{WA}}$ is finally calculated via:
$$
I^{\\text{WA}}_i = 0.1 + 0.9 / \\left( 1 + (I^\\text{WBGT}_i / \\alpha_1)^{\\alpha_2} \\right)
$$
An annual average work ability indiator, $I^{\\text{WA}}$ is calculated via:
$$
I^{\\text{WA}} = \\frac{1}{365} \\sum_{i = 1}^{365} I^{\\text{WA}}_i
$$
The OS-Climate-generated indicators are inferred from CMIP6 data, averaged over 6 models: ACCESS-CM2, CMCC-ESM2, CNRM-CM6-1, MPI-ESM1-2-LR, MIROC6 and NorESM2-MM.
The indicators are generated for periods: 'historical' (averaged over 1995-2014), 2030 (2021-2040), 2040 (2031-2050) and 2050 (2041-2060).

                """
                + methodology_doc,
                "filename": "mean_work_loss_high_{scenario}_{year}",
                "map": {"colormap": "mean_work_loss/high"},
                "units": "fractional loss",
                "scenarios": [
                    {"id": "ssp585", "years": [2030, 2040, 2050]},
                    {"id": "ssp245", "years": [2030, 2040, 2050]},
                    {"id": "historical", "years": [1980]},
                ],
            },
        ]

        wri_riverine_inundation_models = [
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "000000000WATCH",
                "display_name": "WRI/Baseline",
                "description": """
World Resources Institute Aqueduct Floods baseline riverine model using historical data.

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [{"id": "historical", "years": [1980], "periods": [{"year": 1980, "map_id": "gw4vgq"}]}],
            },
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "00000NorESM1-M",
                "display_name": "WRI/NorESM1-M",
                "description": """
World Resources Institute Aqueduct Floods riverine model using GCM model from
Bjerknes Centre for Climate Research, Norwegian Meteorological Institute.

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "0000GFDL-ESM2M",
                "display_name": "WRI/GFDL-ESM2M",
                "description": """
World Resource Institute Aqueduct Floods riverine model using GCM model from
Geophysical Fluid Dynamics Laboratory (NOAA).

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "0000HadGEM2-ES",
                "display_name": "WRI/HadGEM2-ES",
                "description": """
World Resource Institute Aqueduct Floods riverine model using GCM model:
Met Office Hadley Centre.

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "00IPSL-CM5A-LR",
                "display_name": "WRI/IPSL-CM5A-LR",
                "description": """
World Resource Institute Aqueduct Floods riverine model using GCM model from
Institut Pierre Simon Laplace

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "RiverineInundation",
                "path": "riverine_inundation/wri/v2",
                "id": "MIROC-ESM-CHEM",
                "display_name": "WRI/MIROC-ESM-CHEM",
                "description": """World Resource Institute Aqueduct Floods riverine model using
 GCM model from Atmosphere and Ocean Research Institute
 (The University of Tokyo), National Institute for Environmental Studies, and Japan Agency
 for Marine-Earth Science and Technology.

                """
                + aqueduct_description,
                "filename": "inunriver_{scenario}_{id}_{year}",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inunriver_{scenario}_{id}_{year}_rp{return_period:05d}",
                },
                "units": "metres",
                "scenarios": [
                    {
                        "id": "rcp4p5",
                        "years": [2030, 2050, 2080],
                        "periods": [
                            {"year": 2030, "map_id": "ht2kn3"},
                            {"year": 2050, "map_id": "1k4boi"},
                            {"year": 2080, "map_id": "3rok7b"},
                        ],
                    },
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
        ]

        wri_coastal_inundation_models = [
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "nosub",
                "display_name": "WRI/Baseline no subsidence",
                "description": """
World Resources Institute Aqueduct Floods baseline coastal model using historical data. Model excludes subsidence.

                """
                + aqueduct_description,
                "filename": "inuncoast_historical_nosub_hist_0",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_historical_nosub_hist_rp{return_period:04d}_0",
                },
                "units": "metres",
                "scenarios": [{"id": "historical", "years": [1980]}],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "nosub/95",
                "display_name": "WRI/95% no subsidence",
                "description": """
World Resource Institute Aqueduct Floods coastal model, exclusing subsidence; 95th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_nosub_{year}_0",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_nosub_{year}_rp{return_period:04d}_0",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "nosub/5",
                "display_name": "WRI/5% no subsidence",
                "description": """
World Resource Institute Aqueduct Floods coastal model, excluding subsidence; 5th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_nosub_{year}_0_perc_05",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_nosub_{year}_rp{return_period:04d}_0_perc_05",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "nosub/50",
                "display_name": "WRI/50% no subsidence",
                "description": """
World Resource Institute Aqueduct Floods model, excluding subsidence; 50th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_nosub_{year}_0_perc_50",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_nosub_{year}_rp{return_period:04d}_0_perc_50",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "wtsub",
                "display_name": "WRI/Baseline with subsidence",
                "description": """
World Resource Institute Aqueduct Floods model, excluding subsidence; baseline (based on historical data).

                """
                + aqueduct_description,
                "filename": "inuncoast_historical_wtsub_hist_0",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_historical_wtsub_hist_rp{return_period:04d}_0",
                },
                "units": "metres",
                "scenarios": [{"id": "historical", "years": [1980]}],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "wtsub/95",
                "display_name": "WRI/95% with subsidence",
                "description": """
World Resource Institute Aqueduct Floods model, including subsidence; 95th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_wtsub_{year}_0",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_wtsub_{year}_rp{return_period:04d}_0",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "wtsub/5",
                "display_name": "WRI/5% with subsidence",
                "description": """
World Resource Institute Aqueduct Floods model, including subsidence; 5th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_wtsub_{year}_0_perc_05",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_wtsub_{year}_rp{return_period:04d}_0_perc_05",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
            {
                "event_type": "CoastalInundation",
                "path": "coastal_inundation/wri/v2",
                "id": "wtsub/50",
                "display_name": "WRI/50% with subsidence",
                "description": """
World Resource Institute Aqueduct Floods model, including subsidence; 50th percentile sea level rise.

                """
                + aqueduct_description,
                "filename": "inuncoast_{scenario}_wtsub_{year}_0_perc_50",
                "map": {
                    "colormap": "flare_intensity_0_2",
                    "filename": "inuncoast_{scenario}_wtsub_{year}_rp{return_period:04d}_0_perc_50",
                },
                "units": "metres",
                "scenarios": [
                    {"id": "rcp4p5", "years": [2030, 2050, 2080]},
                    {"id": "rcp8p5", "years": [2030, 2050, 2080]},
                ],
            },
        ]

        self.models = osc_chronic_heat_models + wri_riverine_inundation_models + wri_coastal_inundation_models

    def to_hazard_models(self) -> List[HazardModel]:
        models = parse_obj_as(List[HazardModel], self.models)

        # we populate map_id hashes programmatically
        for model in models:

            for scenario in model.scenarios:
                test_periods = scenario.periods
                scenario.periods = []
                for year in scenario.years:
                    name_format = (
                        model.filename if model.map is None or model.map.filename is None else model.map.filename
                    )
                    filename = name_format.format(scenario=scenario.id, year=year, id=model.id, return_period=1000)
                    id = alphanumeric(filename)[0:6]
                    scenario.periods.append(Period(year=year, map_id=id))
                # if a period was specifed explicitly, we check that hash is the same: a build-in check
                if test_periods is not None:
                    for (period, test_period) in zip(scenario.periods, test_periods):
                        if period.map_id != test_period.map_id:
                            raise Exception(
                                f"validation error: hash {period.map_id} different to specified hash {test_period.map_id}"  # noqa: E501
                            )

        return models

    def colormaps(self):
        """Color maps. Key can be identical to a model identifier or more descriptive (if shared by many models)."""
        return colormap_provider.colormaps()


def alphanumeric(text):
    """Return alphanumeric hash from supplied string."""
    hash_int = int.from_bytes(hashlib.sha1(text.encode("utf-8")).digest(), "big")
    return base36encode(hash_int)


def base36encode(number, alphabet="0123456789abcdefghijklmnopqrstuvwxyz"):
    """Converts an integer to a base36 string."""
    if not isinstance(number, int):
        raise TypeError("number must be an integer")

    base36 = ""

    if number < 0:
        raise TypeError("number must be positive")

    if 0 <= number < len(alphabet):
        return alphabet[number]

    while number != 0:
        number, i = divmod(number, len(alphabet))
        base36 = alphabet[i] + base36

    return base36



import logging
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from ..data.hazard_data_provider import (
    get_source_path_osc_chronic_heat,
    get_source_path_wri_coastal_inundation,
    get_source_path_wri_riverine_inundation,
)
from ..data.pregenerated_hazard_model import ZarrHazardModel
from ..models import power_generating_asset_models as pgam
from ..models.chronic_heat_models import ChronicHeatGZN
from ..models.real_estate_models import RealEstateCoastalInundationModel, RealEstateRiverineInundationModel
from ..utils.helpers import get_iterable
from .assets import Asset, IndustrialActivity, PowerGeneratingAsset, RealEstateAsset, TestAsset
from .hazard_event_distrib import HazardEventDistrib
from .hazard_model import HazardDataRequest, HazardDataResponse, HazardModel
from .hazards import ChronicHeat, CoastalInundation, RiverineInundation
from .impact_distrib import ImpactDistrib
from .vulnerability_distrib import VulnerabilityDistrib
from .vulnerability_model import VulnerabilityModelAcuteBase, VulnerabilityModelBase


class AssetImpactResult:
    def __init__(
        self,
        impact: ImpactDistrib,
        vulnerability: VulnerabilityDistrib = None,
        event: HazardEventDistrib = None,
        hazard_data: Iterable[HazardDataResponse] = None,
    ):
        self.impact = impact
        # optional detailed results for drill-dowwn
        self.hazard_data = hazard_data
        self.vulnerability = vulnerability
        self.event = event


def get_default_zarr_source_paths():
    return {
        RiverineInundation: get_source_path_wri_riverine_inundation,
        CoastalInundation: get_source_path_wri_coastal_inundation,
        ChronicHeat: get_source_path_osc_chronic_heat,
    }


def get_default_hazard_model():
    # Model that gets hazard event data from Zarr storage
    return ZarrHazardModel(get_default_zarr_source_paths())


def get_default_vulnerability_models():
    """Get default exposure/vulnerability models for different asset types."""
    return {
        PowerGeneratingAsset: [pgam.InundationModel()],
        RealEstateAsset: [RealEstateCoastalInundationModel(), RealEstateRiverineInundationModel()],
        IndustrialActivity: [ChronicHeatGZN()],
        TestAsset: [pgam.TemperatureModel()],
    }


def calculate_impacts(  # noqa: C901
    assets: Iterable[Asset],
    hazard_model: Optional[HazardModel] = None,
    vulnerability_models: Optional[Any] = None,  #: Optional[Dict[type, Sequence[VulnerabilityModelBase]]] = None,
    *,
    scenario: str,
    year: int,
) -> Dict[Tuple[Asset, type], AssetImpactResult]:
    """Calculate asset level impacts"""
    if hazard_model is None:
        hazard_model = get_default_hazard_model()

    # the models that apply to asset of a particular type
    if vulnerability_models is None:
        vulnerability_models = get_default_vulnerability_models()

    model_assets: Dict[VulnerabilityModelBase, List[Asset]] = defaultdict(
        list
    )  # list of assets to be modelled using vulnerability model
    for asset in assets:
        asset_type = type(asset)
        mappings = vulnerability_models[asset_type]
        for m in mappings:
            model_assets[m].append(asset)

    results = {}

    # as an important performance optimisation, data requests are consolidated for all vulnerability models
    # because different vulnerability models may query the same hazard data sets
    # note that key for request is [vulnerability model, asset]
    assetRequests: Dict[
        Tuple[VulnerabilityModelBase, Asset], Union[HazardDataRequest, Iterable[HazardDataRequest]]
    ] = {}

    logging.info("Generating vulnerability model hazard data requests")
    for model, assets in model_assets.items():
        for asset in assets:
            assetRequests[(model, asset)] = model.get_data_requests(asset, scenario=scenario, year=year)

    logging.info("Retrieving hazard data")
    event_requests = [req for requests in assetRequests.values() for req in get_iterable(requests)]
    responses = hazard_model.get_hazard_events(event_requests)

    logging.info("Calculating impacts")
    for model, assets in model_assets.items():
        logging.info(
            "Applying model {0} to {1} assets of type {2}".format(
                type(model).__name__, len(assets), type(assets[0]).__name__
            )
        )

        for asset in assets:
            requests = assetRequests[(model, asset)]
            hazard_data = [responses[req] for req in get_iterable(requests)]
            if isinstance(model, VulnerabilityModelAcuteBase):
                impact, vul, event = model.get_impact_details(asset, hazard_data)
                results[(asset, model.hazard_type)] = AssetImpactResult(
                    impact, vulnerability=vul, event=event, hazard_data=hazard_data
                )
            elif isinstance(model, VulnerabilityModelBase):
                impact = model.get_impact(asset, hazard_data)
                results[(asset, model.hazard_type)] = AssetImpactResult(impact, hazard_data=hazard_data)

    return results



from typing import List, Union

import numpy as np

# tempting to use typing.get_args
# and use generics for nicer syntax to capture event type, but requires Python 3.8 onwards
# https://stackoverflow.com/questions/48572831/how-to-access-the-type-arguments-of-typing-generic


class VulnerabilityDistrib:
    """Vulnerability distribution as a discrete matrix."""

    __slots__ = ["_event_type", "_intensity_bins", "_impact_bins", "_prob_matrix"]

    def __init__(
        self,
        event_type: type,
        intensity_bins: Union[List[float], np.ndarray],
        impact_bins: Union[List[float], np.ndarray],
        prob_matrix: Union[List[List[float]], np.ndarray],
    ):
        """Create a new vulnerability distribution.
        Args:
            event_type: type of event
            intensity_bins: non-decreasing intensity bin bounds
            impact_bins: non-decreasing impact bin bounds
            prob_matrix: matrix of probabilities with size [len(intensity_bins) - 1, len(impact_bins) - 1]
        """
        self._event_type = event_type
        self._intensity_bins = np.array(intensity_bins)
        self._impact_bins = np.array(impact_bins)
        self._prob_matrix = np.array(prob_matrix)

    # sdef from_mean_impact(intensity_bins: List[float], impact_bins: List[float]):

    @property
    def event_type(self) -> type:
        return self._event_type

    def intensity_bin_bounds(self):
        return zip(self._intensity_bins[0:-1], self._intensity_bins[1:])

    def impact_bin_bounds(self):
        return zip(self._impact_bins[0:-1], self._impact_bins[1:])

    @property
    def impact_bins(self) -> np.ndarray:
        return self._impact_bins

    @property
    def intensity_bins(self) -> np.ndarray:
        return self._intensity_bins

    @property
    def prob_matrix(self) -> np.ndarray:
        return self._prob_matrix



import importlib.resources
import json
from abc import ABC, abstractmethod
from typing import Iterable, List, Tuple, Union

import numpy as np

import physrisk.data.static.vulnerability

from ..api.v1.common import VulnerabilityCurves
from .assets import Asset
from .curve import ExceedanceCurve
from .hazard_event_distrib import HazardEventDistrib
from .hazard_model import HazardDataRequest, HazardDataResponse, HazardEventDataResponse
from .impact_distrib import ImpactDistrib
from .vulnerability_distrib import VulnerabilityDistrib
from .vulnerability_matrix_provider import VulnMatrixProvider

PLUGINS = dict()


def repeat(num_times):
    def decorator_repeat(func):
        ...  # Create and return a wrapper function

    return decorator_repeat


def applies_to_events(event_types):
    def decorator_events(func):
        PLUGINS[func.__name__] = func
        return func

    return decorator_events


def applies_to_assets(asset_types):
    def decorator_events(func):
        PLUGINS[func.__name__] = func
        return func

    return decorator_events


def get_vulnerability_curves_from_resource(id: str) -> VulnerabilityCurves:
    with importlib.resources.open_text(physrisk.data.static.vulnerability, id + ".json") as f:
        curve_set = VulnerabilityCurves(**json.load(f))
        return curve_set


class VulnerabilityModelBase(ABC):
    def __init__(self, model: str, hazard_type: type):
        self.model = model
        self.hazard_type = hazard_type
        self._event_types: List[type] = []
        self._asset_types: List[type] = []

    @abstractmethod
    def get_data_requests(
        self, asset: Asset, *, scenario: str, year: int
    ) -> Union[HazardDataRequest, Iterable[HazardDataRequest]]:
        """Provide the one or more hazard event data requests required in order to calculate
        the VulnerabilityDistrib and HazardEventDistrib for the asset."""
        ...

    @abstractmethod
    def get_impact(self, asset: Asset, event_data: Iterable[HazardDataResponse]) -> ImpactDistrib:
        ...


class VulnerabilityModelAcuteBase(VulnerabilityModelBase):
    """Models generate the VulnerabilityDistrib and HazardEventDistrib of an
    Asset.
    """

    def __init__(self, model: str, hazard_type: type):
        super().__init__(model, hazard_type)

    @abstractmethod
    def get_distributions(
        self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]
    ) -> Tuple[VulnerabilityDistrib, HazardEventDistrib]:
        """Return distributions for asset: VulnerabilityDistrib and HazardEventDistrib.
        The hazard event data is used to do this.

        Args:
            asset: the asset.
            event_data_responses: the responses to the requests made by get_event_data_requests, in the same order.
        """
        ...

    def get_impact(self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]):
        impact, _, _ = self.get_impact_details(asset, event_data_responses)
        return impact

    def get_impact_details(
        self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]
    ) -> Tuple[ImpactDistrib, VulnerabilityDistrib, HazardEventDistrib]:
        """Return impact distribution along with vulnerability and hazard event distributions used to infer this.

        Args:
            asset: the asset.
            event_data_responses: the responses to the requests made by get_event_data_requests, in the same order.
        """
        vulnerability_dist, event_dist = self.get_distributions(asset, event_data_responses)
        impact_prob = vulnerability_dist.prob_matrix.T @ event_dist.prob
        return (
            ImpactDistrib(vulnerability_dist.event_type, vulnerability_dist.impact_bins, impact_prob),
            vulnerability_dist,
            event_dist,
        )

    def _check_event_type(self):
        if self.hazard_type not in self._event_types:
            raise NotImplementedError(f"model does not support events of type {self.hazard_type.__name__}")


class VulnerabilityModel(VulnerabilityModelAcuteBase):
    """A vulnerability model that requires only specification of distributions of impacts for given intensities,
    by implementing get_impact_curve."""

    def __init__(self, *, model: str = "", hazard_type: type, impact_bin_edges):
        self.hazard_type = hazard_type
        self.model = model
        self.impact_bin_edges = impact_bin_edges

    def get_data_requests(
        self, asset: Asset, *, scenario: str, year: int
    ) -> Union[HazardDataRequest, Iterable[HazardDataRequest]]:
        return HazardDataRequest(
            self.hazard_type, asset.longitude, asset.latitude, scenario=scenario, year=year, model=self.model
        )

    def get_distributions(
        self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]
    ) -> Tuple[VulnerabilityDistrib, HazardEventDistrib]:

        (event_data,) = event_data_responses
        assert isinstance(event_data, HazardEventDataResponse)

        intensity_curve = ExceedanceCurve(1.0 / event_data.return_periods, event_data.intensities)
        intensity_bin_edges, probs = intensity_curve.get_probability_bins()
        intensity_bin_centres = (intensity_bin_edges[1:] + intensity_bin_edges[:-1]) / 2

        vul = VulnerabilityDistrib(
            self.hazard_type,
            intensity_bin_edges,
            self.impact_bin_edges,
            self.get_impact_curve(intensity_bin_centres, asset).to_prob_matrix(self.impact_bin_edges),
        )

        event = HazardEventDistrib(self.hazard_type, intensity_bin_edges, probs)
        return vul, event

    @abstractmethod
    def get_impact_curve(self, intensities, asset) -> VulnMatrixProvider:
        """Get the impact curves for the intensities specified."""
        ...


class DeterministicVulnerabilityModel(VulnerabilityModelAcuteBase):
    """A vulnerability model that requires only specification of a damage/disruption curve.
    This simple model contains no uncertainty around damage/disruption."""

    def __init__(self, *, event_type: type, intensities, impacts, model=""):
        self.event_type = event_type
        self.intensities = intensities
        self.impacts = impacts
        self.model = model

    def get_data_requests(
        self, asset: Asset, *, scenario: str, year: int
    ) -> Union[HazardDataRequest, Iterable[HazardDataRequest]]:
        return HazardDataRequest(
            self.event_type, asset.longitude, asset.latitude, scenario=scenario, year=year, model=self.model
        )

    def get_distributions(
        self, asset: Asset, event_data_responses: Iterable[HazardDataResponse]
    ) -> Tuple[VulnerabilityDistrib, HazardEventDistrib]:

        (event_data,) = event_data_responses
        assert isinstance(event_data, HazardEventDataResponse)

        intensity_curve = ExceedanceCurve(1.0 / event_data.return_periods, event_data.intensities)
        intensity_bin_edges, probs = intensity_curve.get_probability_bins()

        impact_bins_edges = np.interp(intensity_bin_edges, self.intensities, self.impacts)

        vul = VulnerabilityDistrib(
            type(self.event_type), intensity_bin_edges, impact_bins_edges, np.eye(len(impact_bins_edges) - 1)
        )
        event = HazardEventDistrib(type(self.event_type), intensity_bin_edges, probs)
        return vul, event



import numpy as np
import scipy.stats as stats

from ..kernel.hazards import RiverineInundation
from ..kernel.vulnerability_matrix_provider import VulnMatrixProvider
from ..kernel.vulnerability_model import VulnerabilityModel, applies_to_events


@applies_to_events([RiverineInundation])
class ExampleCdfBasedVulnerabilityModel(VulnerabilityModel):
    def __init__(self, *, model: str, event_type: type):
        self.intensities = np.array([0, 0.01, 0.5, 1.0, 1.5, 2, 3, 4, 5, 6])
        self.impact_means = np.array([0, 0.2, 0.44, 0.58, 0.68, 0.78, 0.85, 0.92, 0.96, 1.0])
        self.impact_stddevs = np.array([0, 0.17, 0.14, 0.14, 0.17, 0.14, 0.13, 0.10, 0.06, 0])
        impact_bin_edges = np.array([0, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
        super().__init__(model=model, hazard_type=event_type, impact_bin_edges=impact_bin_edges)

    def get_impact_curve(self, intensities, asset):
        # we interpolate the mean and standard deviation and use this to construct distributions
        impact_means = np.interp(intensities, self.intensities, self.impact_means)
        impact_stddevs = np.interp(intensities, self.intensities, self.impact_stddevs)
        return VulnMatrixProvider(
            intensities, impact_cdfs=[checked_beta_distrib(m, s) for m, s in zip(impact_means, impact_stddevs)]
        )


def delta_cdf(y):
    return lambda x: np.where(x >= y, 1, 0)


def checked_beta_distrib(mean, std):
    if mean == 0:
        return delta_cdf(0)
    if mean == 1.0:
        return delta_cdf(1)
    else:
        return beta_distrib(mean, std)


def beta_distrib(mean, std):
    cv = std / mean
    a = (1 - mean) / (cv * cv) - mean
    b = a * (1 - mean) / mean
    return lambda x, a=a, b=b: stats.beta.cdf(x, a, b)



from itertools import chain
from typing import List, Tuple

import numpy as np
import zarr
import zarr.core

from ..utils.lazy_importing import lazy_import

rasterio = lazy_import("rasterio")
tifffile = lazy_import("tifffile")
affine = lazy_import("affine")


def zarr_read(path, longitudes, latitudes):
    """A version that uses Zarr rather than GDAL. Typically faster than GDAL / rasterio."""
    with tifffile.tifffile.Tifffile(path) as tif:
        scale: Tuple[float, float, float] = tif.geotiff_metadata["ModelPixelScale"]
        tie_point: List[float] = tif.geotiff_metadata["ModelTiepoint"]
        store = tif.series[0].aszarr()
        zarray = zarr.open(store, mode="r")
        # shape: List[int] = tif.series[0].shape
        i, j, k, x, y, z = tie_point[0:6]
        sx, sy, sz = scale
        trans = affine.Affine(sx, 0.0, x - i * sx, 0.0, -sy, y + j * sy)
        inv_trans = ~trans
        mat = np.array(inv_trans).reshape(3, 3)
        coords = np.vstack((longitudes, latitudes, np.ones(len(longitudes))))
        frac_image_coords = mat @ coords
        image_coords = np.floor(frac_image_coords).astype(int)
        assert zarray is zarr.core.Array
        data = zarray.get_coordinate_selection((image_coords[1, :], image_coords[0, :]))  # type: ignore
        return data


def dataset_read_bounded(dataset, longitudes, latitudes, window_half_width=0.01):
    hw = window_half_width
    offsets = [[0, 0], [-hw, -hw], [-hw, hw], [hw, hw], [hw, -hw]]
    points = []
    for offset in offsets:
        points = chain(points, [[lon + offset[0], lat + offset[1]] for (lon, lat) in zip(longitudes, latitudes)])

    samples = np.array(list(rasterio.sample.sample_gen(dataset, points)))
    samples.resize([len(offsets), len(longitudes)])
    max_samples = np.max(samples, 0)

    return max_samples


def dataset_read_points(dataset, longitudes, latitudes, window_half_width=0.01):
    points = [[lon, lat] for (lon, lat) in zip(longitudes, latitudes)]
    samples = np.array(list(rasterio.sample.sample_gen(dataset, points)))
    return samples


def dataset_read_windows(dataset, longitudes, latitudes, window_half_width=0.01):
    # seem to need to do one window at a time: potentially slow
    hw = window_half_width
    samples = []
    for (lon, lat) in zip(longitudes, latitudes):
        win = rasterio.windows.from_bounds(
            lon - hw, lat - hw, lon + hw, lat + hw, dataset.transform
        )  # left, bottom, right, top
        max_intensity = np.max(dataset.read(1, window=win))
        samples.append(max_intensity[0])
    return samples


def file_read_bounded(path, longitudes, latitudes, window_half_width=0.01):
    with rasterio.open(path) as dataset:
        return dataset_read_bounded(dataset, longitudes, latitudes, window_half_width)


def file_read_points(path, longitudes, latitudes):
    with rasterio.open(path) as dataset:
        return dataset_read_points(dataset, longitudes, latitudes)



from typing import Optional


class Asset:
    def __init__(self, latitude: float, longitude: float, **kwargs):
        self.latitude = latitude
        self.longitude = longitude
        self.__dict__.update(kwargs)


class PowerGeneratingAsset(Asset):
    pass


class RealEstateAsset(Asset):
    def __init__(self, latitude: float, longitude: float, *, location: str, type: str):
        super().__init__(latitude, longitude)
        self.location = location
        self.type = type


class IndustrialActivity(Asset):
    def __init__(self, latitude: float, longitude: float, *, location: Optional[str] = None, type: str):
        super().__init__(latitude, longitude)
        self.location = location
        self.type = type


class TestAsset(Asset):
    pass



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Sequence

import numpy as np

from .assets import Asset
from .calculation import calculate_impacts, get_default_hazard_model, get_default_vulnerability_models
from .financial_model import FinancialModelBase
from .hazard_model import HazardModel
from .impact_distrib import ImpactDistrib, ImpactType
from .vulnerability_model import VulnerabilityModelAcuteBase


class Aggregator(ABC):
    @abstractmethod
    def get_aggregation_keys(self, asset: Asset, impact: ImpactDistrib) -> List:
        ...


class DefaultAggregator(Aggregator):
    def get_aggregation_keys(self, asset: Asset, impact: ImpactDistrib) -> List:
        return [(impact.hazard_type.__name__), ("root")]


class LossModel:
    def __init__(
        self,
        hazard_model: Optional[HazardModel] = None,
        vulnerability_models: Optional[Dict[type, Sequence[VulnerabilityModelAcuteBase]]] = None,
    ):
        self.hazard_model = get_default_hazard_model() if hazard_model is None else hazard_model
        self.vulnerability_models = (
            get_default_vulnerability_models() if vulnerability_models is None else vulnerability_models
        )

    """Calculates the financial impact on a list of assets."""

    def get_financial_impacts(
        self,
        assets: Sequence[Asset],
        *,
        financial_model: FinancialModelBase,
        scenario: str,
        year: int,
        aggregator: Optional[Aggregator] = None,
        currency: str = "EUR",
        sims: int = 100000
    ):

        if aggregator is None:
            aggregator = DefaultAggregator()

        aggregation_pools: Dict[str, np.ndarray] = {}

        results = calculate_impacts(assets, self.hazard_model, self.vulnerability_models, scenario=scenario, year=year)
        # the impacts in the results are either fractional damage or a fractional disruption

        rg = np.random.Generator(np.random.MT19937(seed=111))

        for (asset, hazard_type), result in results.items():
            # look up keys for results
            impact = result.impact
            keys = aggregator.get_aggregation_keys(asset, impact)
            # transform units of impact into currency for aggregation

            # Monte-Carlo approach: note that if correlations of distributions are simple and model is otherwise linear
            # then calculation by closed-form expression is preferred
            impact_samples = self.uncorrelated_samples(impact, sims, rg)

            if impact.impact_type == ImpactType.damage:
                loss = financial_model.damage_to_loss(asset, impact_samples, currency)
            else:  # impact.impact_type == ImpactType.disruption:
                loss = financial_model.disruption_to_loss(asset, impact_samples, year, currency)

            for key in keys:
                if key not in aggregation_pools:
                    aggregation_pools[key] = np.zeros(sims)
                aggregation_pools[key] += loss  # type: ignore

        measures = {}
        percentiles = [0, 10, 20, 40, 60, 80, 90, 95, 97.5, 99, 99.5, 99.9]
        for key, loss in aggregation_pools.items():
            measures[key] = {
                "percentiles": percentiles,
                "percentile_values": np.percentile(loss, percentiles),
                "mean": np.mean(loss),
            }

        return measures

    def uncorrelated_samples(self, impact: ImpactDistrib, samples: int, generator: np.random.Generator) -> np.ndarray:
        return impact.to_exceedance_curve().get_samples(generator.uniform(size=samples))



import os
from pathlib import PurePosixPath
from typing import Callable, MutableMapping, Optional

import numpy as np
import s3fs
import zarr
from affine import Affine


def get_env(key: str, default: Optional[str] = None) -> str:
    value = os.environ.get(key)
    if value is None:
        if default is not None:
            return default
        raise ValueError(f"environment variable {key} not present")
    else:
        return value


class ZarrReader:
    """Reads hazard event data from Zarr files, including OSC-format-specific attributes."""

    # environment variable names:
    __access_key = "OSC_S3_ACCESS_KEY"
    __secret_key = "OSC_S3_SECRET_KEY"
    __S3_bucket = "OSC_S3_BUCKET"  # e.g. redhat-osc-physical-landing-647521352890 on staging
    __zarr_path = "OSC_S3_HAZARD_PATH"  # e.g. hazard/hazard.zarr on staging

    def __init__(
        self,
        store: Optional[MutableMapping] = None,
        path_provider: Optional[Callable[..., str]] = None,
        get_env: Callable[[str, Optional[str]], str] = get_env,
    ):
        """Create a ZarrReader.

        Args:
            store: if not supplied, create S3Map store.
            path_provider: function that provides path to the data set based on a set ID.
            get_env: allows override obtaining of environment variables.
        """
        if store is None:
            # if no store is provided, attempt to connect to an S3 bucket
            if get_env is None:
                raise TypeError("if no store specified, get_env is required to provide credentials")

            access_key = get_env(self.__access_key, None)
            secret_key = get_env(self.__secret_key, None)
            s3_bucket = get_env(self.__S3_bucket, "redhat-osc-physical-landing-647521352890")
            zarr_path = get_env(self.__zarr_path, "hazard/hazard.zarr")

            s3 = s3fs.S3FileSystem(anon=False, key=access_key, secret=secret_key)

            store = s3fs.S3Map(
                root=str(PurePosixPath(s3_bucket, zarr_path)),
                s3=s3,
                check=False,
            )

        self._root = zarr.open(store, mode="r")
        self._path_provider = path_provider
        pass

    def get_curves(self, set_id, longitudes, latitudes, interpolation="floor"):
        """Get intensity curve for each latitude and longitude coordinate pair.

        Args:
            set_id: string or tuple representing data set, converted into path by path_provider.
            longitudes: list of longitudes.
            latitudes: list of latitudes.
            interpolation: interpolation method, "floor", "linear", "max" or "min".

        Returns:
            curves: numpy array of intensity (no. coordinate pairs, no. return periods).
            return_periods: return periods in years.
        """

        # assume that calls to this are large, not chatty
        if len(longitudes) != len(latitudes):
            raise ValueError("length of longitudes and latitudes not equal")
        path = self._path_provider(set_id) if self._path_provider is not None else set_id
        z = self._root[path]  # e.g. inundation/wri/v2/<filename>

        # OSC-specific attributes contain tranform and return periods
        t = z.attrs["transform_mat3x3"]  # type: ignore
        transform = Affine(t[0], t[1], t[2], t[3], t[4], t[5])

        # in the case of acute risks, index_values will contain the return periods
        index_values = z.attrs.get("index_values", [0])
        if index_values is None:
            index_values = [0]

        image_coords = self._get_coordinates(longitudes, latitudes, transform)

        if interpolation == "floor":
            image_coords = np.floor(image_coords).astype(int)
            iz = np.tile(np.arange(z.shape[0]), image_coords.shape[1])  # type: ignore
            iy = np.repeat(image_coords[1, :], len(index_values))
            ix = np.repeat(image_coords[0, :], len(index_values))

            data = z.get_coordinate_selection((iz, iy, ix))  # type: ignore
            return data.reshape([len(longitudes), len(index_values)]), np.array(index_values)

        elif interpolation in ["linear", "max", "min"]:
            res = ZarrReader._linear_interp_frac_coordinates(z, image_coords, index_values, interpolation=interpolation)
            return res, np.array(index_values)

        else:
            raise ValueError("interpolation must have value 'floor', 'linear', 'max' or 'min")

    def get_max_curves(self, set_id, longitudes, latitudes, interpolation="floor", delta_km=1.0, n_grid=5):
        """Get maximal intensity curve for a grid around a given latitude and longitude coordinate pair.

        Args:
            set_id: string or tuple representing data set, converted into path by path_provider.
            longitudes: list of longitudes.
            latitudes: list of latitudes.
            interpolation: interpolation method, "floor", "linear", "max" or "min".
            delta_km: linear distance in kilometres of the side of the square grid surrounding a given position.
            n_grid: number of grid points along the latitude and longitude dimensions used for
                    calculating the maximal value.

        Returns:
            curves_max: numpy array of maximum intensity on the grid for a given coordinate pair
            (no. coordinate pairs, no. return periods).
            return_periods: return periods in years.
        """
        KILOMETRES_PER_DEGREE = 110.574
        n_data = len(latitudes)
        delta_deg = delta_km / KILOMETRES_PER_DEGREE
        grid = np.linspace(-0.5, 0.5, n_grid)
        lats_grid_baseline = np.broadcast_to(latitudes.reshape((n_data, 1, 1)), (len(latitudes), n_grid, n_grid))
        lons_grid_baseline = np.broadcast_to(longitudes.reshape((n_data, 1, 1)), (len(longitudes), n_grid, n_grid))
        lats_grid_offsets = delta_deg * grid.reshape((1, n_grid, 1))
        lons_grid_offsets = (
            delta_deg * grid.reshape((1, 1, n_grid)) / (np.cos((np.pi / 180) * latitudes).reshape(n_data, 1, 1))
        )
        lats_grid = lats_grid_baseline + lats_grid_offsets
        lons_grid = lons_grid_baseline + lons_grid_offsets
        curves_, return_periods = self.get_curves(
            set_id, lons_grid.reshape(-1), lats_grid.reshape(-1), interpolation=interpolation
        )
        curves_max = np.max(curves_.reshape((n_data, n_grid * n_grid, len(return_periods))), axis=1)
        return curves_max, return_periods

    @staticmethod
    def _linear_interp_frac_coordinates(z, image_coords, return_periods, interpolation="linear"):
        """Return linear interpolated data from fractional row and column coordinates."""
        icx = np.floor(image_coords[0, :]).astype(int)[..., None]
        # note periodic boundary condition
        ix = np.concatenate([icx, icx, (icx + 1) % z.shape[2], (icx + 1) % z.shape[2]], axis=1)[..., None].repeat(
            len(return_periods), axis=2
        )  # points, 4, return_periods

        icy = np.floor(image_coords[1, :]).astype(int)[..., None]
        iy = np.concatenate([icy, icy + 1, icy, icy + 1], axis=1)[..., None].repeat(len(return_periods), axis=2)

        iz = (
            np.arange(len(return_periods), dtype=int)[None, ...]
            .repeat(4, axis=0)[None, ...]
            .repeat(image_coords.shape[1], axis=0)
        )

        data = z.get_coordinate_selection((iz, iy, ix))  # type: ignore # index, row, column

        NAN_VALUE = -9999.0

        if interpolation == "linear":
            xf = image_coords[0, :][..., None] - icx  # type: ignore
            yf = image_coords[1, :][..., None] - icy  # type: ignore
            w0 = (1 - yf) * (1 - xf)
            w1 = yf * (1 - xf)
            w2 = (1 - yf) * xf
            w3 = yf * xf
            w = np.transpose(np.array([w0, w1, w2, w3]), (1, 0, 2))
            mask = 1 - np.isnan(np.where(data == NAN_VALUE, np.nan, data))
            w_good = w * mask
            w_good_sum = np.transpose(
                np.sum(w_good, axis=1).reshape(tuple([1]) + np.sum(w_good, axis=1).shape), axes=(1, 0, 2)
            )
            w_used = np.divide(w_good, np.where(w_good_sum == 0.0, np.nan, w_good_sum))
            return np.nan_to_num(np.sum(w_used * data, axis=1), nan=NAN_VALUE)

        elif interpolation == "max":
            data = np.where(data == NAN_VALUE, -np.inf, data)
            return np.nan_to_num(
                np.maximum.reduce([data[:, 0, :], data[:, 1, :], data[:, 2, :], data[:, 3, :]]),
                nan=NAN_VALUE,
                neginf=NAN_VALUE,
            )

        elif interpolation == "min":
            data = np.where(data == NAN_VALUE, np.inf, data)
            return np.nan_to_num(
                np.minimum.reduce([data[:, 0, :], data[:, 1, :], data[:, 2, :], data[:, 3, :]]),
                nan=NAN_VALUE,
                posinf=NAN_VALUE,
            )

        else:
            raise ValueError("interpolation must have value 'linear', 'max' or 'min")

    @staticmethod
    def _get_coordinates(longitudes, latitudes, transform: Affine):
        coords = np.vstack((longitudes, latitudes, np.ones(len(longitudes))))  # type: ignore
        inv_trans = ~transform
        mat = np.array(inv_trans).reshape(3, 3)
        frac_image_coords = mat @ coords
        return frac_image_coords



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Mapping, Protocol, Tuple

import numpy as np


class HazardDataRequest:
    """Request for hazard data. The event_type determines whether the hazard is acute or chronic.
    An acute hazard is an event and the response will therefore comprise hazard intensities for the
    different event return periods. A chronic hazard on the other hand is a shift in a climate parameter
    and the parameter value is returned."""

    def __init__(self, hazard_type: type, longitude: float, latitude: float, *, model: str, scenario: str, year: int):
        """Create HazardDataRequest.

        Args:
            event_type: type of hazard.
            longitude: required longitude.
            latitude: required latitude.
            model: model identifier.
            scenario: identifier of scenario, e.g. rcp8p5 (RCP 8.5).
            year: projection year, e.g. 2080.
        """
        self.hazard_type = hazard_type
        self.longitude = longitude
        self.latitude = latitude
        self.model = model
        self.scenario = scenario
        self.year = year

    def group_key(self):
        """Key used to group EventDataRequests into batches."""
        return tuple((self.hazard_type, self.model, self.scenario, self.year))


class HazardDataResponse:
    parameter: np.ndarray


class HazardEventDataResponse(HazardDataResponse):
    """Response to HazardDataRequest for acute hazards."""

    def __init__(self, return_periods: np.ndarray, intensities: np.ndarray):
        """Create HazardEventDataResponse.

        Args:
            return_periods: return periods in years.
            intensities: hazard event intensity for each return period.
        """
        self.return_periods = return_periods
        self.intensities = intensities


class HazardParameterDataResponse(HazardDataResponse):
    """Response to HazardDataRequest."""

    def __init__(self, parameter: np.ndarray):
        """Create HazardParameterDataResponse.

        Args:
            parameter: the chronic hazard parameter value.
        """
        self.parameter = parameter


class HazardModel(ABC):
    """Hazard event model. The model accepts a set of EventDataRequests and returns the corresponding
    EventDataResponses."""

    @abstractmethod
    def get_hazard_events(self, requests: List[HazardDataRequest]) -> Mapping[HazardDataRequest, HazardDataResponse]:
        """Process the hazard data requests and return responses."""
        ...


class DataSource(Protocol):
    def __call__(self, longitudes, latitudes, *, model: str, scenario: str, year: int) -> Tuple[np.ndarray, np.ndarray]:
        ...


class CompositeHazardModel(HazardModel):
    """Hazard Model that uses other models to process EventDataRequests."""

    def __init__(self, hazard_models: Dict[type, HazardModel]):
        self.hazard_models = hazard_models

    def get_hazard_events(self, requests: List[HazardDataRequest]) -> Mapping[HazardDataRequest, HazardDataResponse]:
        requests_by_event_type = defaultdict(list)

        for request in requests:
            requests_by_event_type[request.hazard_type].append(request)

        responses: Dict[HazardDataRequest, HazardDataResponse] = {}
        for event_type, reqs in requests_by_event_type.items():
            events_reponses = self.hazard_models[event_type].get_hazard_events(reqs)
            responses.update(events_reponses)

        return responses



from .assets import Asset, PowerGeneratingAsset
from .calculation import calculate_impacts
from .curve import ExceedanceCurve
from .hazard_event_distrib import HazardEventDistrib
from .hazards import Drought, Hazard, RiverineInundation
from .vulnerability_distrib import VulnerabilityDistrib
from .vulnerability_model import VulnerabilityModelAcuteBase



import importlib.resources
import json
from typing import Dict, List, Union

import numpy as np

import physrisk.data.static
from physrisk.api.v1.common import Countries, Country


def get_countries_from_resource():
    with importlib.resources.open_text(physrisk.data.static, "world.json") as f:
        countries = Countries(**json.load(f))
        return dict((c.country, c) for c in countries.items)


def get_countries_json():
    """Get countries and continents, populating json."""

    import geopandas as gpd

    world = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))  # type: ignore
    countries = world[["continent", "name", "iso_a3"]]

    countries = [
        Country(continent=continent, country=country, country_iso_a3=code)
        for (continent, country, code) in zip(world["continent"], world["name"], world["iso_a3"])
    ]

    return json.dumps(Countries(items=countries).dict(), sort_keys=True, indent=4)


def get_countries_and_continents(longitudes: Union[List[float], np.ndarray], latitudes: Union[List[float], np.ndarray]):
    """Only for use when on-boarding; look up country and continent (e.g. for use in vulnerability models) by
    latitude and longitude."""

    # Geopandas draws in a number of libraries, including GDAL, so we probably(?)
    # want to confine its use to pre-processing / on-boarding of data
    # In particular, country/continent look-up is probably something to do pre-onboarding

    import geopandas as gpd

    # consider using map here https://gadm.org/download_world.html
    world = gpd.read_file(gpd.datasets.get_path("naturalearth_lowres"))  # type: ignore
    gdf = gpd.GeoDataFrame(crs=world.crs, geometry=gpd.points_from_xy(longitudes, latitudes))
    result = gpd.sjoin(gdf, world, how="left")

    return list(result["name"]), list(result["continent"])


class World:

    countries: Dict[str, Country] = get_countries_from_resource()



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict

import numpy as np

from .assets import Asset


class FinancialDataProvider(ABC):
    @abstractmethod
    def get_asset_value(self, asset: Asset, currency: str) -> float:
        """Return the current value of the asset in specified currency."""
        ...

    @abstractmethod
    def get_asset_aggregate_cashflows(self, asset: Asset, start: datetime, end: datetime, currency: str) -> float:
        """Return the expected sum of the cashflows generated by the Asset between start and end, in
        specified currency."""
        ...


class FinancialModelBase(ABC):
    @abstractmethod
    def damage_to_loss(self, asset: Asset, impact: np.ndarray, currency: str):
        """Convert the fractional damage of the specified asset to a financial loss."""
        ...

    @abstractmethod
    def disruption_to_loss(self, asset: Asset, impact: np.ndarray, year: int, currency: str):
        """Convert the fractional annual disruption of the specified asset to a financial loss."""
        ...


class FinancialModel(FinancialModelBase):
    """ "Financial Model using a FinancialDataProvider as source of information."""

    def __init__(self, data_provider: FinancialDataProvider):
        self.data_provider = data_provider

    def damage_to_loss(self, asset: Asset, impact: np.ndarray, currency: str):
        return self.data_provider.get_asset_value(asset, currency) * impact

    def disruption_to_loss(self, asset: Asset, impact: np.ndarray, year: int, currency: str):
        return (
            self.data_provider.get_asset_aggregate_cashflows(
                asset, datetime(year, 1, 1), datetime(year, 12, 31), currency
            )
            * impact
        )


class CompositeFinancialModel(FinancialModelBase):
    """Financial model split by asset type."""

    def __init__(self, financial_models: Dict[type, FinancialModelBase]):
        self.financial_models = financial_models

    def damage_to_loss(self, asset: Asset, impact: np.ndarray, currency: str):
        return self.financial_models[type(asset)].damage_to_loss(asset, impact, currency)

    def disruption_to_loss(self, asset: Asset, impact: np.ndarray, year: int, currency: str):
        return self.financial_models[type(asset)].disruption_to_loss(asset, impact, year, currency)



from typing import Iterable, Union

import numpy as np
from scipy.stats import norm

from physrisk.kernel.assets import Asset
from physrisk.kernel.hazard_model import HazardDataRequest, HazardDataResponse
from physrisk.kernel.hazards import ChronicHeat
from physrisk.kernel.impact_distrib import ImpactDistrib, ImpactType
from physrisk.kernel.vulnerability_model import VulnerabilityModelBase


class ChronicHeatGZN(VulnerabilityModelBase):
    """Model which estiamtes the labour productivity impact based on chronic heat based on the paper "Neidell M,
    Graff Zivin J, Sheahan M,  Willwerth J, Fant C, Sarofim M, et al. (2021) Temperature and work:
    Time allocated to work under varying climate and labor market conditions."
    Average annual work hours are based on USA values reported by the OECD for 2021."""

    def __init__(self, model: str = "mean_degree_days_above_32c", delta=True):
        super().__init__(model, ChronicHeat)  # opportunity to give a model hint, but blank here
        self.time_lost_per_degree_day = 4.671  # This comes from the paper converted to celsius
        self.time_lost_per_degree_day_se = 2.2302  # This comes from the paper converted to celsius
        self.total_labour_hours = 107460
        self.delta = delta

    def get_data_requests(
        self, asset: Asset, *, scenario: str, year: int
    ) -> Union[HazardDataRequest, Iterable[HazardDataRequest]]:
        """Request the hazard data needed by the vulnerability model for a specific asset
        (this is a Google-style doc string)

        Args:
            asset: Asset for which data is requested.
            scenario: Climate scenario of calculation.
            year: Projection year of calculation.

        Returns:
            Single data requests.
        """

        # specify hazard data needed. Model string is hierarchical and '/' separated.
        model = "mean_degree_days/above/32c"

        return [
            HazardDataRequest(
                self.hazard_type,
                asset.longitude,
                asset.latitude,
                scenario="historical",
                year=1980,
                model=model,
            ),
            HazardDataRequest(
                self.hazard_type,
                asset.longitude,
                asset.latitude,
                scenario=scenario,
                year=year,
                model=model,
            ),
        ]

    def get_impact(self, asset: Asset, data_responses: Iterable[HazardDataResponse]) -> ImpactDistrib:
        """Calcaulate impact (disruption) of asset based on the hazard data returned.

        Args:
            asset: Asset for which impact is calculated.
            data_responses: responses to the hazard data requests generated in get_data_requests.

        Returns:
            Probability distribution of impacts.
        """

        baseline_dd_above_mean, scenario_dd_above_mean = data_responses

        # assert isinstance(baseline_dd_above_mean, HazardParameterDataResponse)
        # assert isinstance(scenario_dd_above_mean, HazardParameterDataResponse)

        delta_dd_above_mean: float = scenario_dd_above_mean.parameter - baseline_dd_above_mean.parameter * self.delta

        hours_worked = self.total_labour_hours
        fraction_loss_mean = (delta_dd_above_mean * self.time_lost_per_degree_day) / hours_worked
        fraction_loss_std = (delta_dd_above_mean * self.time_lost_per_degree_day_se) / hours_worked

        return get_impact_distrib(fraction_loss_mean, fraction_loss_std, ChronicHeat, ImpactType.disruption)


def get_impact_distrib(
    fraction_loss_mean: float, fraction_loss_std: float, hazard_type: type, impact_type: ImpactType
) -> ImpactDistrib:
    """Calculate impact (disruption) of asset based on the hazard data returned.

    Args:
        fraction_loss_mean: mean of the impact distribution
        fraction_loss_std: standard deviation of the impact distribution
        hazard_type: Hazard Type.
        impact_type: Impact Type.

    Returns:
        Probability distribution of impacts.
    """
    impact_bins = np.concatenate(
        [
            np.linspace(-0.001, 0.001, 1, endpoint=False),
            np.linspace(0.001, 0.01, 9, endpoint=False),
            np.linspace(0.01, 0.1, 10, endpoint=False),
            np.linspace(0.1, 0.999, 10, endpoint=False),
            np.linspace(0.999, 1.001, 2),
        ]
    )

    probs_cumulative = np.vectorize(lambda x: norm.cdf(x, loc=fraction_loss_mean, scale=max(1e-12, fraction_loss_std)))(
        impact_bins
    )
    probs_cumulative[-1] = np.maximum(probs_cumulative[-1], 1.0)
    probs = np.diff(probs_cumulative)

    probs_norm = np.sum(probs)
    prob_differential = 1 - probs_norm
    if probs_norm < 1e-8:
        if fraction_loss_mean <= 0.0:
            probs = np.concatenate((np.array([1.0]), np.zeros(len(impact_bins) - 2)))
        elif fraction_loss_mean >= 1.0:
            probs = np.concatenate((np.zeros(len(impact_bins) - 2), np.array([1.0])))
    else:
        probs[0] = probs[0] + prob_differential

    return ImpactDistrib(hazard_type, impact_bins, probs, impact_type)



from typing import List, Optional, Union

from pydantic import BaseModel

from physrisk.api.v1.common import IntensityCurve


class MapInfo(BaseModel):
    """Provides information about map layer"""

    colormap: str
    filename: Optional[str]


class Period(BaseModel):
    """A period belonging to a scenario"""

    year: int
    map_id: str


class Scenario(BaseModel):
    """Scenario ID and the list of available years for that scenario e.g. RCP8.5 = 'rcp8.5'"""

    id: str
    years: List[int]
    periods: Optional[List[Period]]


class HazardModel(BaseModel):
    """Provides the scenarios associated with a hazard model."""

    event_type: str
    path: str
    id: str
    display_name: str
    description: str
    filename: str
    map: Optional[MapInfo]
    scenarios: List[Scenario]
    units: str


# region HazardAvailability


class HazardEventAvailabilityRequest(BaseModel):
    event_types: Union[List[str], None]  # e.g. RiverineInundation


class HazardEventAvailabilityResponse(BaseModel):
    models: List[HazardModel]
    colormaps: dict


# endregion

# region HazardEventData


class HazardEventDataRequestItem(BaseModel):
    longitudes: List[float]
    latitudes: List[float]
    request_item_id: str
    event_type: str  # e.g. RiverineInundation
    model: str
    scenario: str  # e.g. rcp8p5
    year: int


class HazardEventDataRequest(BaseModel):
    interpolation: str = "floor"
    items: List[HazardEventDataRequestItem]


class HazardEventDataResponseItem(BaseModel):
    intensity_curve_set: List[IntensityCurve]
    request_item_id: str
    event_type: str
    model: str
    scenario: str
    year: int


class HazardEventDataResponse(BaseModel):
    items: List[HazardEventDataResponseItem]


# endregion



from collections import defaultdict
from typing import Dict, Tuple

import numpy as np
import scipy.stats as stats

from physrisk.api.v1.common import VulnerabilityCurve, VulnerabilityCurves
from physrisk.kernel.assets import RealEstateAsset
from physrisk.kernel.vulnerability_model import VulnerabilityModel, VulnMatrixProvider

from ..kernel.hazards import CoastalInundation, RiverineInundation
from ..kernel.vulnerability_model import applies_to_events, get_vulnerability_curves_from_resource


class RealEstateInundationModel(VulnerabilityModel):
    _default_impact_bin_edges = np.array([0, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    _default_resource = "EU JRC global flood depth-damage functions"

    def __init__(
        self,
        *,
        event_type: type,
        model: str,
        resource: str = _default_resource,
        impact_bin_edges=_default_impact_bin_edges
    ):
        """
        Inundation vulnerability model for real estates assets. Applies to both riverine and coastal inundation.

        Args:
            event_type: Event type.
            model: optional identifier for hazard event model, passed to HazardModel.
            resource: embedded resource identifier used to infer vulnerability matrix.
            impact_bin_edges: specifies the impact (fractional damage/disruption bins).
        """

        curve_set: VulnerabilityCurves = get_vulnerability_curves_from_resource(resource)

        # for this model, key for looking up curves is (location, asset_type), e.g. ('Asian', 'Building/Industrial')
        self.vulnerability_curves = dict(((c.location, c.asset_type), c) for c in curve_set.items)
        self.vuln_curves_by_type = defaultdict(list)
        self.proxy_curves: Dict[Tuple[str, str], VulnerabilityCurve] = {}
        for item in curve_set.items:
            self.vuln_curves_by_type[item.asset_type].append(item)

        # global circulation parameter 'model' is a hint; can be overriden by hazard model
        super().__init__(model=model, hazard_type=event_type, impact_bin_edges=impact_bin_edges)

    def get_impact_curve(self, intensities, asset: RealEstateAsset):
        # we interpolate the mean and standard deviation and use this to construct distributions
        # assert asset is RealEstateAsset

        key = (asset.location, asset.type)
        curve = self.vulnerability_curves[key]

        std_curve = curve
        if len(curve.impact_std) == 0:
            if key not in self.proxy_curves:
                self.proxy_curves[key] = self.closest_curve_of_type(curve, asset)
            std_curve = self.proxy_curves[key]

        impact_means = np.interp(intensities, curve.intensity, curve.impact_mean)
        impact_stddevs = np.interp(intensities, std_curve.intensity, std_curve.impact_std)

        return VulnMatrixProvider(
            intensities, impact_cdfs=[checked_beta_distrib(m, s) for m, s in zip(impact_means, impact_stddevs)]
        )

    def closest_curve_of_type(self, curve: VulnerabilityCurve, asset: RealEstateAsset):
        # we return the standard deviations of the damage curve most similar to the asset location
        candidate_set = list(cand for cand in self.vuln_curves_by_type[asset.type] if (len(cand.impact_std) > 0))
        sum_square_diff = (self.sum_square_diff(curve, cand) for cand in candidate_set)
        lowest = np.argmin(np.array(list(sum_square_diff)))
        return candidate_set[lowest]

    def sum_square_diff(self, curve1: VulnerabilityCurve, curve2: VulnerabilityCurve):
        return np.sum((curve1.impact_mean - np.interp(curve1.intensity, curve2.intensity, curve2.impact_mean)) ** 2)


def delta_cdf(y):
    return lambda x: np.where(x >= y, 1, 0)


def checked_beta_distrib(mean, std):
    if mean == 0:
        return delta_cdf(0)
    if mean == 1.0:
        return delta_cdf(1)
    else:
        return beta_distrib(mean, std)


def beta_distrib(mean, std):
    cv = std / mean
    a = (1 - mean) / (cv * cv) - mean
    b = a * (1 - mean) / mean
    return lambda x, a=a, b=b: stats.beta.cdf(x, a, b)


@applies_to_events([CoastalInundation])
class RealEstateCoastalInundationModel(RealEstateInundationModel):
    def __init__(
        self,
        *,
        model: str = "wtsub/95",
        resource: str = RealEstateInundationModel._default_resource,
        impact_bin_edges=RealEstateInundationModel._default_impact_bin_edges
    ):
        # by default include subsidence and 95% sea-level rise
        super().__init__(
            event_type=CoastalInundation, model=model, resource=resource, impact_bin_edges=impact_bin_edges
        )


@applies_to_events([RiverineInundation])
class RealEstateRiverineInundationModel(RealEstateInundationModel):
    def __init__(
        self,
        *,
        model: str = "MIROC-ESM-CHEM",
        resource: str = RealEstateInundationModel._default_resource,
        impact_bin_edges=RealEstateInundationModel._default_impact_bin_edges
    ):
        # by default request HazardModel to use "MIROC-ESM-CHEM" GCM
        super().__init__(
            event_type=RiverineInundation, model=model, resource=resource, impact_bin_edges=impact_bin_edges
        )



from typing import List, Union

import numpy as np


def add_x_value_to_curve(x, curve_x, curve_y):
    """Add an x value to a curve, interpolated from the existing curve.
    curve_x and curve_y are the curve x and y values.
    curve_x is sorted non-decreasing. This function may be used to align curves and bins.
    """
    # note some care needed as multiple identical curve_x are permitted: cannot simply use np.interp
    # this might be a candidate for numba optimization

    i = np.searchsorted(curve_x, x)

    if i == len(curve_y):
        curve_y = np.insert(curve_y, i, curve_y[i - 1])  # flat extrapolation
        curve_x = np.insert(curve_x, i, x)
    elif x == curve_x[i]:
        # point already exists; nothing to do
        return curve_x, curve_y
    elif i == 0:
        curve_y = np.insert(curve_y, 0, curve_y[0])  # flat extrapolation
        curve_x = np.insert(curve_x, 0, x)
    else:
        pl, pu = curve_y[i - 1], curve_y[i]
        il, iu = curve_x[i - 1], curve_x[i]
        # linear interpolation; quadratic interpolation (linear in probability density) may also be of interest
        prob = pl + (x - il) * (pu - pl) / (iu - il)
        curve_y = np.insert(curve_y, i, prob)
        curve_x = np.insert(curve_x, i, x)

    return curve_x, curve_y


def to_exceedance_curve(bin_edges, probs):
    """An exceedance curve gives the probability that the random variable is greater than the value,
    a type of cumulative probability.
    """
    nz = np.asarray(probs > 0).nonzero()
    fnz = nz[0][0] if len(nz[0]) > 0 else 0
    nz_values = bin_edges[fnz:]
    nz_probs = probs[fnz:]
    cum_prob = np.insert(np.cumsum(nz_probs[::-1]), 0, 0.0)[::-1]
    return ExceedanceCurve(cum_prob, nz_values)


def process_bin_edges_and_probs(bin_edges, probs, range_fraction=0.05):
    r = bin_edges[-1] - bin_edges[0]
    r = bin_edges[0] if r == 0 else r
    new_edges = []
    new_probs = []
    # say we have edges
    # 0, 1, 2, 3, 3, 3, 5, 7
    # we want to convert to
    # 0, 1, 2, 3 + d, 5, 7
    # 2d = (7 - 0) * 0.01
    i = 0
    while i < len(bin_edges):
        j = __next_non_equal_index(bin_edges, i)
        if j == i + 1:
            i = i + 1
            new_edges.append(bin_edges[i])
            new_probs.append(probs[i])
            continue
        if j >= len(bin_edges):
            delta = r * range_fraction
        else:
            delta = min(r * range_fraction, 0.25 * (bin_edges[j] - bin_edges[i]))
        new_edges.append(bin_edges[i])
        new_edges.append(bin_edges[i] + delta)
        new_probs.append(np.sum(probs[i:j]))
        i = j
    return new_edges, new_probs


def process_bin_edges_for_graph(bin_edges, range_fraction=0.05):
    """Process infinitessimal (zero width) bins for graph display.
    We make width 5% of range or 1/4 the width to the next bin edge, whichever is smaller
    """
    r = bin_edges[-1] - bin_edges[0]
    r = bin_edges[0] if r == 0 else r
    new_edges = np.copy(bin_edges)
    # say we have edges
    # 0, 1, 2, 3, 3, 3, 5, 7
    # we want to convert to
    # 0, 1, 2, 3, 3 + d, 3 + 2d, 5, 7
    # 2d = (7 - 0) * 0.01
    i = 0
    while i < len(bin_edges):
        j = __next_non_equal_index(bin_edges, i)
        if j == i + 1:
            i = i + 1
            continue
        if j >= len(bin_edges):
            delta = r * range_fraction / (j - i - 1)
        else:
            delta = min(r * range_fraction, 0.25 * (bin_edges[j] - bin_edges[i])) / (j - i - 1)
        offset = delta
        for k in range(i + 1, j):
            new_edges[k] = new_edges[k] + offset
            offset += delta
        i = j
    return new_edges


def __next_non_equal_index(ndarray, i):
    j = i + 1
    c = ndarray[i]
    while j < len(ndarray) and ndarray[j] == c:
        j = j + 1
    return j


class ExceedanceCurve:
    """A point on an exceedance curve comprises an value, v, and a probability, p.
    p is the probability that the random variable >= v, e.g. an event occurs with event value (e.g. intensity) >= v."""

    __slots__ = ["probs", "values"]

    def __init__(self, probs: Union[List[float], np.ndarray], values: Union[List[float], np.ndarray]):
        """Create a new asset event distribution.
        Args:
            probs: exceedance probabilities (must be sorted and decreasing).
            values: values (must be sorted and non-decreasing).
        """

        # probabilities must be sorted and decreasing
        # values must be sorted and non-decreasing (intens[i + 1] >= intens[i])
        if len(probs) != len(values):
            raise ValueError("same number of probabilities and values expected")
        if not np.all(np.diff(probs) <= 0):
            raise ValueError("probs must be sorted and decreasing")
        if not np.all(np.diff(values) >= 0):
            raise ValueError("values must be sorted and non-decreasing")

        self.probs = np.array(probs)
        self.values = np.array(values)

    def add_value_point(self, value):
        """Add a point to the curve with specified value and exceedance
        probability determined from existing curve by linear interpolation.
        """

        values, probs = add_x_value_to_curve(value, self.values, self.probs)
        return ExceedanceCurve(probs, values)

    def get_value(self, prob):
        return np.interp(prob, self.probs[::-1], self.values[::-1])

    def get_probability_bins(self):
        r"""Convert from exceedance (cumulative) probability to bins of constant probability.
        This is equivalent to the assumption of linear interpolation of exceedance points.

        .. math::
            p^\text{b}_i = p^\text{e}_{i + 1} - p^\text{e}_i

        Returns:
            value_bins (ndarray), probs: the contiguous bin lower and upper values, probabilities of each bin
            If value_bins is of lenth n then ther are n-1 bins and n-1 probabilities

        """
        # value bins are contiguous
        value_bins = self.values[:]
        probs = self.probs[:-1] - self.probs[1:]  # type: ignore
        return value_bins, probs

    def get_samples(self, uniforms):
        """Return value, v, for each probability p in uniforms such that p is the probability that the random variable
        < v."""
        return np.where(uniforms > (1.0 - self.probs[0]), np.interp(uniforms, 1.0 - self.probs, self.values), 0.0)



from pydantic import BaseModel

from physrisk.api.v1.common import Assets


class ExamplePortfoliosRequest(BaseModel):
    """Example portfolios request."""


class ExamplePortfoliosResponse(BaseModel):
    """Example portfolios response."""

    assets: Assets



import os

from setuptools import setup


def get_version():
    dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(dir, "VERSION"), "r") as fp:
        return fp.read()


setup(version=get_version(), include_package_data=True)



from enum import Enum
from typing import List, Union

import numpy as np

from .curve import to_exceedance_curve


class ImpactType(Enum):
    damage = 1
    disruption = 2


class ImpactDistrib:
    """Impact distributions specific to an asset."""

    __slots__ = ["__hazard_type", "__impact_bins", "__prob", "impact_type"]

    def __init__(
        self,
        hazard_type: type,
        impact_bins: Union[List[float], np.ndarray],
        prob: Union[List[float], np.ndarray],
        impact_type: ImpactType = ImpactType.damage,
    ):
        """Create a new asset event distribution.
        Args:
            event_type: type of event
            impact_bins: non-decreasing impact bin bounds
            prob: probabilities with size [len(intensity_bins) - 1]
        """
        self.__hazard_type = hazard_type
        self.__impact_bins = np.array(impact_bins)
        self.impact_type = impact_type
        self.__prob = np.array(prob)

    def impact_bins_explicit(self):
        return zip(self.__impact_bins[0:-1], self.__impact_bins[1:])

    def mean_impact(self):
        return np.sum((self.__impact_bins[:-1] + self.__impact_bins[1:]) * self.__prob / 2)

    def to_exceedance_curve(self):
        return to_exceedance_curve(self.__impact_bins, self.__prob)

    @property
    def hazard_type(self) -> type:
        return self.__hazard_type

    @property
    def impact_bins(self) -> np.ndarray:
        return self.__impact_bins

    @property
    def prob(self) -> np.ndarray:
        return self.__prob



from typing import List, Union

import numpy as np


class Distribution:
    def __init__(self, mean, std_dev):
        self.mean = mean
        self.std_dev = std_dev


class VulnMatrixProvider:
    """Provides the impact on an asset, an impact being either a fractional damage or disruption,
    that occurs as a result of a hazard event of a given intensity."""

    __slots__ = ["intensities", "impact_cdfs"]

    def __init__(
        self,
        intensities: Union[List[float], np.ndarray],
        impact_cdfs=None,
    ):
        """Create a new asset event distribution.

        Args:
            intensities: possible intensities of hazard event.
            impacts: fractional damage or fractional average disruption occurring as a result
                of hazard event of given intensity.
            distributions: provides the pdf and optiononally cdf of the impact distribution
        """

        if not np.all(np.diff(intensities) >= 0):
            raise ValueError("intensities must be sorted and increasing")

        self.intensities = np.array(intensities)
        self.impact_cdfs = impact_cdfs

    def to_prob_matrix(self, impact_bin_edges):
        # construct a cdf probability matrix at each intensity point
        # the probability is the prob that the impact is greater than the specified
        cdf_matrix = np.empty([len(self.intensities), len(impact_bin_edges)])

        for i, _ in enumerate(self.intensities):
            cdf_matrix[i, :] = self.impact_cdfs[i](impact_bin_edges)  # type: ignore

        prob_matrix = cdf_matrix[:, 1:] - cdf_matrix[:, :-1]

        return prob_matrix



import os
from abc import ABC
from typing import List, MutableMapping, Optional

from typing_extensions import Protocol

from .zarr_reader import ZarrReader


class SourcePath(Protocol):
    """Provides path to hazard event data source. Each source should have its own implementation.
    Args:
        model: model identifier.
        scenario: identifier of scenario, e.g. rcp8p5 (RCP 8.5).
        year: projection year, e.g. 2080.
    """

    def __call__(self, *, model: str, scenario: str, year: int) -> str:
        ...


class HazardDataProvider(ABC):
    def __init__(
        self,
        get_source_path: SourcePath,
        *,
        store: Optional[MutableMapping] = None,
        zarr_reader: Optional[ZarrReader] = None,
        interpolation: Optional[str] = "floor",
    ):
        """Create an EventProvider.

        Args:
            get_source_path: provides the path to the hazard event data source depending on year/scenario/model.
        """
        self._get_source_path = get_source_path
        self._reader = zarr_reader if zarr_reader is not None else ZarrReader(store=store)
        if interpolation not in ["floor", "linear"]:
            raise ValueError("interpolation must be 'floor' or 'linear'")
        self._interpolation = interpolation


class AcuteHazardDataProvider(HazardDataProvider):
    """Provides hazard event intensities for a single Hazard (type of hazard event)."""

    def __init__(
        self,
        get_source_path: SourcePath,
        *,
        store: Optional[MutableMapping] = None,
        zarr_reader: Optional[ZarrReader] = None,
        interpolation: Optional[str] = "floor",
    ):
        super().__init__(get_source_path, store=store, zarr_reader=zarr_reader, interpolation=interpolation)

    def get_intensity_curves(
        self, longitudes: List[float], latitudes: List[float], *, model: str, scenario: str, year: int
    ):
        """Get intensity curve for each latitude and longitude coordinate pair.

        Args:
            longitudes: list of longitudes.
            latitudes: list of latitudes.
            model: model identifier.
            scenario: identifier of scenario, e.g. rcp8p5 (RCP 8.5).
            year: projection year, e.g. 2080.

        Returns:
            curves: numpy array of intensity (no. coordinate pairs, no. return periods).
            return_periods: return periods in years.
        """

        path = self._get_source_path(model=model, scenario=scenario, year=year)
        curves, return_periods = self._reader.get_curves(
            path, longitudes, latitudes, self._interpolation
        )  # type: ignore
        return curves, return_periods


class ChronicHazardDataProvider(HazardDataProvider):
    """Provides hazard parameters for a single type of chronic hazard."""

    def __init__(
        self,
        get_source_path: SourcePath,
        *,
        store: Optional[MutableMapping] = None,
        zarr_reader: Optional[ZarrReader] = None,
        interpolation: Optional[str] = "floor",
    ):
        super().__init__(get_source_path, store=store, zarr_reader=zarr_reader, interpolation=interpolation)

    def get_parameters(self, longitudes: List[float], latitudes: List[float], *, model: str, scenario: str, year: int):
        """Get hazard parameters for each latitude and longitude coordinate pair.

        Args:
            longitudes: list of longitudes.
            latitudes: list of latitudes.
            model: model identifier.
            scenario: identifier of scenario, e.g. rcp8p5 (RCP 8.5).
            year: projection year, e.g. 2080.

        Returns:
            parameters: numpy array of parameters
        """

        path = self._get_source_path(model=model, scenario=scenario, year=year)
        parameters, _ = self._reader.get_curves(path, longitudes, latitudes, self._interpolation)
        return parameters[:, 0]


# region World Resource Aqueduct Model


def _wri_inundation_prefix():
    return "inundation/wri/v2"


_percentiles_map = {"95": "0", "5": "0_perc_05", "50": "0_perc_50"}
_subsidence_set = {"wtsub", "nosub"}


def get_source_path_wri_coastal_inundation(*, model: str, scenario: str, year: int):
    type = "coast"
    # model is expected to be of the form subsidence/percentile, e.g. wtsub/95
    # if percentile is omitted then 95th percentile is used
    model_components = model.split("/")
    sub = model_components[0]
    if sub not in _subsidence_set:
        raise ValueError("expected model input of the form {subsidence/percentile}, e.g. wtsub/95, nosub/5, wtsub/50")
    perc = "95" if len(model_components) == 1 else model_components[1]
    return os.path.join(
        _wri_inundation_prefix(), f"inun{type}_{cmip6_scenario_to_rcp(scenario)}_{sub}_{year}_{_percentiles_map[perc]}"
    )


def get_source_path_wri_riverine_inundation(*, model: str, scenario: str, year: int):
    type = "river"
    return os.path.join(_wri_inundation_prefix(), f"inun{type}_{cmip6_scenario_to_rcp(scenario)}_{model}_{year}")


def cmip6_scenario_to_rcp(scenario: str):
    """Convention is that CMIP6 scenarios are expressed by identifiers:
    SSP1-2.6: 'ssp126'
    SSP2-4.5: 'ssp245'
    SSP5-8.5: 'ssp585' etc.
    Here we translate to form
    RCP-4.5: 'rcp4p5'
    RCP-8.5: 'rcp8p5' etc.
    """
    if scenario == "ssp126":
        return "rcp2p6"
    elif scenario == "ssp245":
        return "rcp4p5"
    elif scenario == "ssp585":
        return "rcp8p5"
    else:
        if scenario not in ["rcp2p6" "rcp4p5", "rcp8p5", "historical"]:
            raise ValueError(f"unexpcted scneario {scenario}")
        return scenario


# endregion

# region OS-C Chronic Heat Model


def _osc_chronic_heat_prefix():
    return "chronic_heat/osc/v1"


def get_source_path_osc_chronic_heat(*, model: str, scenario: str, year: int):
    type, *levels = model.split("/")

    if type == "mean_degree_days":
        assert levels[0] in ["above", "below"]  # above or below
        assert levels[1] in ["18c", "32c"]  # threshold temperature
        return _osc_chronic_heat_prefix() + "/" + f"{type}_{levels[0]}_{levels[1]}_{scenario}_{year}"

    elif type == "mean_work_loss":
        assert levels[0] in ["low", "medium", "high"]  # work intensity
        return _osc_chronic_heat_prefix() + "/" + f"{type}_{levels[0]}_{scenario}_{year}"

    else:
        raise ValueError("valid types are {valid_types}")


# endregion
